/*
  ktracker (c) 2006 Kris Beevers

  This file is part of ktracker.

  ktracker is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  ktracker is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with ktracker; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

  $Id: tracker.cpp,v 1.1.1.1 2006/10/10 20:41:29 beevek Exp $
*/

#include <inttypes.h>
#include <math.h>
#include <vector>
#include <map>
#include <string>
#include <opencv/cv.h>
#include <opencv/highgui.h>
#include <opencv/cvaux.hpp>
#include "video_input.hpp"
#include "options.hpp"
#include "tracker.hpp"


extern bool gui; // in ktracker

namespace tracker
{

  typedef CvPoint2D32f point_t; // float x, y

  // class that tracks one object matching each input template.  the
  // input templates are images representing typical colors of the
  // object, e.g., green, orange, yellow, etc.  so really, this is
  // just colored blob tracking for a variety of colors.
  class template_tracker
  {
  public:
    template_tracker() : hsv(0), hue(0), backproject(0), first_frame(true) {}
    ~template_tracker();

    // add a "template" image, i.e., create a histogram based on the
    // template image
    bool add_template(const char *tpl, uint8_t label);

    // track the objects for all the templates
    bool track_all_objects(const IplImage *cvimg);

    // get the centroids of the tracked object with the given template
    // label
    void get_centroid(uint8_t label, point_t *c);

    // draw the box determined by the tracker for the given label
    void draw_outline(uint8_t label, IplImage *img, CvScalar color);

  protected:

    struct object_template
    {
      uint8_t label;
      CvHistogram *hist;
      CvRect window;
      CvBox2D box;
      CvConnectedComp comp;
    };

    IplImage *hsv, *hue, *backproject;
    std::vector<object_template> templates;
    bool first_frame;
    uint32_t find_template(uint8_t label) const;
    void make_hue_image(const IplImage *bgr);
  };

  template_tracker::~template_tracker()
  {
    if(hsv) cvReleaseImage(&hsv);
    if(hue) cvReleaseImage(&hue);
    if(backproject) cvReleaseImage(&backproject);
  }

  void template_tracker::make_hue_image(const IplImage *bgr)
  {
    if(!hsv || hsv->width != bgr->width || hsv->height != bgr->height) {
      cvReleaseImage(&hsv);
      hsv = cvCreateImage(cvGetSize(bgr), 8, 3);
    }
    if(!hue || hue->width != bgr->width || hue->height != bgr->height) {
      cvReleaseImage(&hue);
      hue = cvCreateImage(cvGetSize(bgr), 8, 1);
    }
    cvCvtColor(bgr, hsv, CV_BGR2HSV);
    cvSplit(hsv, hue, 0, 0, 0);
  }

  bool template_tracker::add_template(const char *tpl, uint8_t label)
  {
    templates.push_back(object_template());
    object_template &t = templates.back();
    t.label = label;
    IplImage *timg = cvLoadImage(tpl);
    if(!timg)
      return false;

    make_hue_image(timg);

    static int hdims = 40;
    static float hranges_arr[] = {30,180}; // determined experimentally with dragonfly-6mm
    static float *hranges = hranges_arr;
    t.hist = cvCreateHist(1, &hdims, CV_HIST_ARRAY, &hranges, 1);
    cvCalcHist(&hue, t.hist);
    return true;
  }

  bool template_tracker::track_all_objects(const IplImage *cvimg)
  {
    for(uint32_t i = 0; i < templates.size(); ++i) {
      object_template &t = templates[i];
      if(first_frame)
	t.window = cvRect(0, 0, cvimg->width, cvimg->height);

      make_hue_image(cvimg);
      if(!backproject)
	backproject = cvCreateImage(cvGetSize(cvimg), 8, 1);

      cvCalcBackProject(&hue, backproject, t.hist);
      cvCamShift(backproject, t.window,
		 cvTermCriteria(CV_TERMCRIT_EPS | CV_TERMCRIT_ITER, 10, 1),
		 &t.comp, &t.box);
      t.window = t.comp.rect;
      t.box.angle = -t.box.angle;

      // if we lose the object, reset the search window
      if(t.window.width == 0 || t.window.height == 0) {
	t.window = cvRect(0, 0, cvimg->width, cvimg->height);
	t.comp.rect = t.window;
      }
    }
    first_frame = false;
    return true;
  }

  uint32_t template_tracker::find_template(uint8_t label) const
  {
    for(uint32_t i = 0; i < templates.size(); ++i)
      if(templates[i].label == label)
	return i;
    return templates.size();
  }

  void template_tracker::get_centroid(uint8_t label, point_t *c)
  {
    uint32_t idx = find_template(label);
    if(idx >= templates.size())
      return;
    *c = templates[idx].box.center;
  }

  void template_tracker::draw_outline(uint8_t label, IplImage *img, CvScalar color)
  {
    uint32_t idx = find_template(label);
    if(idx >= templates.size())
      return;

    point_t pts[4];
    cvBoxPoints(templates[idx].box, pts);

    cvLine(img, cvPointFrom32f(pts[0]), cvPointFrom32f(pts[1]), color);
    cvLine(img, cvPointFrom32f(pts[1]), cvPointFrom32f(pts[2]), color);
    cvLine(img, cvPointFrom32f(pts[2]), cvPointFrom32f(pts[3]), color);
    cvLine(img, cvPointFrom32f(pts[3]), cvPointFrom32f(pts[0]), color);
  }


  ////////////////////////////////////////////////////////////////////
  // robot tracking stuff for our particular application
  ////////////////////////////////////////////////////////////////////

  enum { FRONT = 0, BACK = 1 };

  bool no_overlays = false;
  uint32_t imgsize, w, h, bpp, grayscale_size;
  template_tracker track;
  IplImage *img_in = 0, *img_out = 0;
  CvMat *intrinsics, *distortion, *udmap_x, *udmap_y, *p_img2ground, *p_ground2img;

  bool load_calibration_file(const char *file)
  {
    intrinsics = cvCreateMat(3, 3, CV_32FC1);
    distortion = cvCreateMat(1, 4, CV_32FC1);
    cvZero(intrinsics);
    cvZero(distortion);

    FILE *in = fopen(file, "r");
    if(!in)
      goto error;

    double v;

    // read intrinsic parameters
    if(fscanf(in, "%lg", &v) != 1) goto error; cvmSet(intrinsics, 0, 0, v); // fx
    if(fscanf(in, "%lg", &v) != 1) goto error; cvmSet(intrinsics, 1, 1, v); // fy
    if(fscanf(in, "%lg", &v) != 1) goto error; cvmSet(intrinsics, 0, 2, v); // cx
    if(fscanf(in, "%lg", &v) != 1) goto error; cvmSet(intrinsics, 1, 2, v); // cy
    cvmSet(intrinsics, 2, 2, 1);

    // read distortion parameters
    for(uint32_t i = 0; i < 4; ++i) {
      if(fscanf(in, "%lg", &v) != 1) goto error;
      cvmSet(distortion, 0, i, v);
    }

    // read perspective transformation matrices for transforming
    // between image and ground plane and vice versa
    p_img2ground = cvCreateMat(3, 3, CV_32FC1);
    p_ground2img = cvCreateMat(3, 3, CV_32FC1);
    for(uint32_t i = 0; i < 3; ++i)
      for(uint32_t j = 0; j < 3; ++j) {
	if(fscanf(in, "%lg", &v) != 1) goto error;
	cvmSet(p_img2ground, i, j, v);
      }
    for(uint32_t i = 0; i < 3; ++i)
      for(uint32_t j = 0; j < 3; ++j) {
	if(fscanf(in, "%lg", &v) != 1) goto error;
	cvmSet(p_ground2img, i, j, v);
      }

    fclose(in);
    return true;
  error:
    if(in) fclose(in);
    cvReleaseMat(&intrinsics);
    cvReleaseMat(&distortion);
    return false;
  }

  void perspective_transform(const point_t &src, point_t *dest, const CvMat *xform)
  {
    CvMat src_mat = cvMat(1, 1, CV_32FC2, (void *)&src);
    CvMat dest_mat = cvMat(1, 1, CV_32FC2, dest);
    cvPerspectiveTransform(&src_mat, &dest_mat, xform);
  }

  bool init(const video_input &input)
  {
    no_overlays = options::quickget<bool>("no-overlays");

    imgsize = input.imgsize;
    w = input.width;
    h = input.height;
    bpp = input.bpp;
    grayscale_size = imgsize / bpp / 8;

    img_in = cvCreateImageHeader(cvSize(w,h), IPL_DEPTH_8U, 3);
    img_in->origin = 0;
    img_out = cvCreateImage(cvSize(w,h), IPL_DEPTH_8U, 3);

    // load camera calibration parameters and initialize the
    // undistortion map
    if(!load_calibration_file(options::quickget<std::string>("calibration-file").c_str()))
      return false;
    udmap_x = cvCreateMat(h, w, CV_32FC1);
    udmap_y = cvCreateMat(h, w, CV_32FC1);
    cvInitUndistortMap(intrinsics, distortion, udmap_x, udmap_y);
    cvReleaseMat(&intrinsics);
    cvReleaseMat(&distortion);

    std::string fn = options::quickget<std::string>("front-template");
    if(!track.add_template(fn.c_str(), FRONT))
      return false;
    fn = options::quickget<std::string>("back-template");
    return track.add_template(fn.c_str(), BACK);
  }

  void cleanup()
  {
    if(img_in)
      cvReleaseImageHeader(&img_in);
    if(img_out)
      cvReleaseImage(&img_out);
    cvReleaseMat(&udmap_x);
    cvReleaseMat(&udmap_y);
    cvReleaseMat(&p_img2ground);
    cvReleaseMat(&p_ground2img);
  }

  // modifies img (if there's a gui)
  void track_robot(uint8_t *img, robot_t &pose)
  {
    // make an IplImage
    cvSetData(img_in, img, w*3);

    // undistort image using the camera intrinsics
    cvRemap(img_in, img_out, udmap_x, udmap_y);

    // then call track_all_objects
    if(!track.track_all_objects(img_out)) {
      printf("error tracking!\n");
      return;
    }

    // and call get_centroid for both front and back
    point_t cent_back, cent_front, g_cent_back, g_cent_front;
    track.get_centroid(FRONT, &cent_front);
    track.get_centroid(BACK, &cent_back);

    // next project the centroids from the image plane onto the ground
    // plane using the calibrated perspective transform
    perspective_transform(cent_front, &g_cent_front, p_img2ground);
    perspective_transform(cent_back, &g_cent_back, p_img2ground);

    // compute the robot centroid (x,y) and orientation t as angle
    // from back to front centroid

    // FIXME how to compute robot centroid?  here's a temporary
    // solution but it might be best to take as input the distance
    // from //front and back centroids to the robot centroid, and then
    // compute //according to front/back centroids, then average
    pose.x = (g_cent_front.x + g_cent_back.x) / 2.0;
    pose.y = (g_cent_front.y + g_cent_back.y) / 2.0;
    pose.t = atan2(g_cent_front.y - g_cent_back.y,
		   g_cent_front.x - g_cent_back.x);


    // if there's a gui, draw boxes around each of the detected
    // objects, dots for robot centroids, and lines for robot
    // orientations
    if(gui && !no_overlays) {
      track.draw_outline(FRONT, img_out, CV_RGB(0,255,0));
      track.draw_outline(BACK, img_out, CV_RGB(255,96,0));
      cvCircle(img_out, cvPointFrom32f(cent_front), 2, CV_RGB(0,255,0), CV_FILLED);
      cvCircle(img_out, cvPointFrom32f(cent_back), 2, CV_RGB(255,96,0), CV_FILLED);

      // project centroid and orientation line endpoint back onto the
      // image plane and render them
      point_t img_robot_cent, img_orient;
      perspective_transform(cvPoint2D32f(pose.x, pose.y), &img_robot_cent, p_ground2img);
      perspective_transform(cvPoint2D32f(pose.x+cos(pose.t), pose.y+sin(pose.t)),
			    &img_orient, p_ground2img);
      cvCircle(img_out, cvPointFrom32f(img_robot_cent), 4, CV_RGB(255,0,255), CV_FILLED);
      cvLine(img_out, cvPointFrom32f(img_robot_cent), cvPointFrom32f(img_orient), CV_RGB(255,0,255));

      memcpy(img, img_out->imageData, imgsize);
    }
  }

}; // namespace tracker
