/*
  ktracker (c) 2006 Kris Beevers

  This file is part of ktracker.

  ktracker is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  ktracker is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with ktracker; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

  $Id: video_input.cpp,v 1.1.1.1 2006/10/10 20:41:29 beevek Exp $
*/

#include "video_input.hpp"
#include <string.h>

#ifdef _HAVE_METEOR
#include <sys/fcntl.h>
#include <sys/mman.h>
#endif

video_input * autoopen_video(const char *filename)
{
  video_input *in = 0;
  if(strstr(filename, "/dev/video1394")) {
#ifdef _HAVE_DRAGONFLY
    in = new video_input_dragonfly;
#else
    in = 0;
#endif
  } else if(strstr(filename, "/dev/video")) {
#ifdef _HAVE_FFMPEG
    in = new video_input_v4l;
#else
    in = 0;
#endif
  } else if(strstr(filename, "/dev/dv1394")) {
#ifdef _HAVE_FFMPEG
    in = new video_input_dv1394;
#else
    in = 0;
#endif
  } else if(strstr(filename, "/dev/meteor")) {
#ifdef _HAVE_METEOR
    in = new video_input_meteor;
#else
    in = 0;
#endif
  } else {
#ifdef _HAVE_FFMPEG
    in = new video_input_file;
#else
    in = 0;
#endif
  }
  if(in && !in->open(filename)) {
    delete in;
    in = 0;
  }
  return in;
}

#ifdef _HAVE_FFMPEG

bool video_input_file::open(const char *input)
{
  // initialize video input
  av_register_all();

  if(do_av_open_input_file(input) != 0) {
    fprintf(stderr, "Can't open %s\n", input);
    return false;
  }
  if(av_find_stream_info(fmtcontext) < 0) {
    fprintf(stderr, "Can't read stream info from %s\n", input);
    return false;
  }
  stream = -1;
  for(int i = 0; i < fmtcontext->nb_streams; ++i)
    if(fmtcontext->streams[i]->codec->codec_type == CODEC_TYPE_VIDEO) {
      stream = i;
      break;
    }
  if(stream == -1) {
    fprintf(stderr, "No video stream in %s\n", input);
    return false;
  }
  cdccontext = fmtcontext->streams[stream]->codec;
  codec = avcodec_find_decoder(cdccontext->codec_id);
  if(codec == 0) {
    fprintf(stderr, "Can't find decoder for %s\n", input);
    return 1;
  }
  if(codec->capabilities & CODEC_CAP_TRUNCATED)
    cdccontext->flags |= CODEC_CAP_TRUNCATED;
  if(avcodec_open(cdccontext, codec) < 0) {
    fprintf(stderr, "Can't open codec for %s\n", input);
    return 1;
  }
  frame = avcodec_alloc_frame();
  frame_bgr = avcodec_alloc_frame();

  width = cdccontext->width;
  height = cdccontext->height;
  bpp = 24;
  imgsize = avpicture_get_size(PIX_FMT_BGR24, width, height);
  imgbuf = new unsigned char[imgsize];
  avpicture_fill((AVPicture *)frame_bgr, imgbuf, PIX_FMT_BGR24, width, height);
  packet.data = 0;
  init_okay = true;
  return true;
}

// file-based input
int video_input_file::do_av_open_input_file(const char *input)
{
  return av_open_input_file(&fmtcontext, input, 0, 0, 0);
}

// video4linux input
int video_input_v4l::do_av_open_input_file(const char *input)
{
  AVFormatParameters format_params;
  format_params.device = input; // should probably be "/dev/video0"
  format_params.channel = 0;
  format_params.standard = "ntsc";
  //  format_params.width = 640;
  //  format_params.height = 480;
  // FIXME FIXME: should have a way to specify resolution
format_params.width = 352;
format_params.height = 288;
  format_params.time_base.den = 29;
  format_params.time_base.num = 1;
  AVInputFormat *iformat = av_find_input_format("video4linux");
  return av_open_input_file(&fmtcontext, "", iformat, 0, &format_params);
}

// dv1394 input
int video_input_dv1394::do_av_open_input_file(const char *input)
{
  AVFormatParameters format_params;
  format_params.device = input; // should probably be "/dev/dv1394"
  AVInputFormat *iformat = av_find_input_format("dv1394");
  return av_open_input_file(&fmtcontext, "", iformat, 0, &format_params);
}

void video_input_file::close()
{
  if(!init_okay)
    return;
  delete [] imgbuf;
  av_free(frame_bgr);
  av_free(frame);
  avcodec_close(cdccontext);
  av_close_input_file(fmtcontext);
  init_okay = false;
}

uint8_t * video_input_file::get_next_frame()
{
  if(imgbuf == 0)
    return 0;

  int frame_finished;
  while(av_read_frame(fmtcontext, &packet) >= 0) {
    if(packet.stream_index != stream)
      continue;
    avcodec_decode_video(cdccontext, frame, &frame_finished,
			 packet.data, packet.size);
    if(frame_finished > 0)
      return frame2bgr();
  }
  return 0;
}

uint8_t * video_input_file::frame2bgr()
{
  img_convert((AVPicture *)frame_bgr, PIX_FMT_BGR24, (AVPicture *)frame,
	      cdccontext->pix_fmt, cdccontext->width, cdccontext->height);
  return frame_bgr->data[0];
}

#endif // _HAVE_FFMPEG


#ifdef _HAVE_METEOR

video_input_meteor::video_input_meteor()
  : video_input(), rawsize(640*480*4), frame(0), frame_bgr(0), fd(-1)
{
  width = 640;
  height = 480;
  imgsize = 640*480*3;
  bpp = 24;
}

bool video_input_meteor::open(const char *input)
{
  fd = open(input, O_RDONLY);
  if(fd < 0)
    return false;
  struct meteor_geomet geo;
  geo.rows = 640;
  geo.columns = 480;
  geo.frames = 1;
  geo.oformat = METEOR_GEO_RGB24;
  if(ioctl(fd, METEORSETGEO, &geo) < 0)
    return false;
  int cmd = METEOR_FMT_NTSC;
  if(ioctl(fd, METEORSFMT, &cmd) < 0)
    return false;
  cmd = METEOR_INPUT_DEV0;
  if(ioctl(fd, METEORSINPUT, &cmd) < 0)
    return false;
#ifdef _NO_MEMMAP
  frame = new uint8_t[rawsize];
#else
  frame = (uint8_t *)mmap(0, rawsize, PROT_READ, MAP_SHARED, fp, 0);
#endif
  frame_bgr = new uint8_t[imgsize];
  init_okay = true;
  return true;
}

void video_input_meteor::close()
{
  if(!init_okay)
    return;
  delete [] frame_bgr;
#ifdef _NO_MEMMAP
  delete [] frame;
#endif
  close(fd);
  init_okay = false;
}

uint8_t * video_input_meteor::get_next_frame()
{
  if(!init_okay)
    return 0;
#ifdef _NO_MEMMAP
  if(read(fd, frame, rawsize) < rawsize)
    return 0;
#else
  static int cmd = METEOR_CAP_SINGLE;
  if(ioctl(fp, METEORCAPTUR, &cmd) < 0)
    return 0;
#endif
  return frame2bgr();
}

uint8_t * video_input_meteor::frame2bgr()
{
  uint32_t i = 0, j = 0;
  for(; i < rawsize; ++i) { // ++i skips the extra null byte in the raw data
    frame_bgr[j++] = frame[i++];
    frame_bgr[j++] = frame[i++];
    frame_bgr[j++] = frame[i++];
  }
  return frame_bgr;
}

#endif // _HAVE_HETEOR

#ifdef _HAVE_DRAGONFLY

video_input_dragonfly::video_input_dragonfly()
  : video_input(), frame_bgr(0)
{
  width = 640;
  height = 480;
  imgsize = 640*480*3;
  bpp = 24;
}

bool video_input_dragonfly::open(const char *input)
{
  handle = dc1394_create_handle(0); // assume card 0
  if(!handle)
    return false;
  int ret;
  nodeid_t *nodes = dc1394_get_camera_nodes(handle, &ret, 1);
  if(ret < 1) // no cameras found
    goto error;
  ret = dc1394_dma_setup_capture
    (handle, nodes[0], 0, FORMAT_VGA_NONCOMPRESSED, MODE_640x480_MONO,
     SPEED_400, FRAMERATE_30, 8 /* num buffers */, 1 /* drop frames */,
     input, &camera);
  dc1394_free_camera_nodes(nodes);
  if(ret != DC1394_SUCCESS)
    goto error;
  if(dc1394_start_iso_transmission(handle, camera.node) != DC1394_SUCCESS)
    goto error_dma;

  frame_bgr = new uint8_t[imgsize];
  init_okay = true;
  return true;

 error_dma:
  dc1394_dma_unlisten(handle, &camera);
  dc1394_dma_release_camera(handle, &camera);
 error:
  dc1394_destroy_handle(handle);
  return false;
}

void video_input_dragonfly::close()
{
  if(!init_okay)
    return;
  delete [] frame_bgr;
  dc1394_stop_iso_transmission(handle, camera.node);
  dc1394_dma_unlisten(handle, &camera);
  dc1394_dma_release_camera(handle, &camera);
  raw1394_destroy_handle(handle);
  init_okay = false;
}

uint8_t * video_input_dragonfly::get_next_frame()
{
  if(!init_okay)
    return 0;

  if(dc1394_dma_single_capture(&camera) != DC1394_SUCCESS)
    return 0;
  frame2bgr();
  dc1394_dma_done_with_buffer(&camera);

  return frame_bgr;
}

uint8_t * video_input_dragonfly::frame2bgr()
{
  // see bayer.cpp
  dragonfly_bayer2bgr((uint8_t *)camera.capture_buffer, frame_bgr, width, height);
  return frame_bgr;
}

#endif // _HAVE_DRAGONFLY
