/*
  ktracker (c) 2006 Kris Beevers

  This file is part of ktracker.

  ktracker is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  ktracker is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with ktracker; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

  $Id: video_input.hpp,v 1.1.1.1 2006/10/10 20:41:29 beevek Exp $
*/

#ifndef _VIDEO_INPUT_HPP
#define _VIDEO_INPUT_HPP

#include <inttypes.h>

#ifdef _HAVE_FFMPEG
#include <ffmpeg/avcodec.h>
#include <ffmpeg/avformat.h>
#endif
#ifdef _HAVE_METEOR
#include <dev/bktr/ioctl_meteor.h>
#endif
#ifdef _HAVE_DRAGONFLY
#include <libraw1394/raw1394.h>
#include <libdc1394/dc1394_control.h>
#endif

class video_input
{
public:
  video_input() : init_okay(false) {}
  virtual ~video_input() {}
  virtual bool open(const char *input) = 0;
  virtual void close() = 0;
  virtual uint8_t * get_next_frame() = 0; // returns frame in BGR24 format

  uint32_t width, height, bpp, imgsize;
protected:
  bool init_okay;
};

#ifdef _HAVE_FFMPEG
// file (e.g. avi, mpeg, etc.) input
class video_input_file : public video_input
{
public:
  video_input_file() : video_input() {}
  virtual ~video_input_file() { close(); }
  bool open(const char *input);
  void close();
  uint8_t * get_next_frame();

protected:
  virtual int do_av_open_input_file(const char *input);
  uint8_t * frame2bgr();

  AVFormatContext *fmtcontext;
  AVCodecContext *cdccontext;
  AVCodec *codec;
  AVFrame *frame, *frame_bgr;
  int stream;
  uint8_t *imgbuf;

  AVPacket packet;
  int bytes_remaining;
  unsigned char *raw_data;
};

// video4linux input
class video_input_v4l : public video_input_file
{
public:
  video_input_v4l() : video_input_file() {}
  ~video_input_v4l() { close(); }
protected:
  int do_av_open_input_file(const char *input);
};

// dv1394 (firewire) input, e.g., a camcorder
class video_input_dv1394 : public video_input_file
{
public:
  video_input_dv1394() : video_input_file() {}
  ~video_input_dv1394() { close(); }
protected:
  int do_av_open_input_file(const char *input);
};
#endif // _HAVE_FFMPEG

#ifdef _HAVE_METEOR
class video_input_meteor : public video_input
{
public:
  video_input_meteor();
  ~video_input_meteor() { close(); }
  bool open(const char *input);
  void close();
  uint8_t * get_next_frame();

protected:
  uint8_t * frame2bgr();
  uint32_t rawsize;
  uint8_t *frame, *frame_bgr;
  int fd;
};
#endif // _HAVE_METEOR

#ifdef _HAVE_DRAGONFLY
// useful because, while ffmpeg supports dc1394, it doesn't support
// transfer in the right format (MONO) and doesn't do Bayer decoding
class video_input_dragonfly : public video_input
{
public:
  video_input_dragonfly();
  ~video_input_dragonfly() { close(); }
  bool open(const char *input);
  void close();
  uint8_t * get_next_frame();

protected:
  uint8_t * frame2bgr();
  uint8_t *frame_bgr;
  raw1394handle_t handle;
  dc1394_cameracapture camera;
};

// bayer decoding, in bayer.cpp
void dragonfly_bayer2bgr(const uint8_t *bayer, uint8_t *bgr, uint32_t sx, uint32_t sy);

#endif // _HAVE_DRAGONFLY

// picks the right subclass based on the filename; returns 0 if we
// don't know how to open the filename
video_input * autoopen_video(const char *filename);

#endif // _VIDEO_INPUT_HPP
